from dataclasses import dataclass
from typing import List, Optional

from elasticsearch_reindex.const import DEFAULT_CHECK_INTERVAL, DEFAULT_CONCURRENT_TASKS


@dataclass
class Index:
    """
    Dataclass for storing ES index cat data.
    """

    name: str
    docs_count: int


@dataclass
class Config:
    """
    Dataclass for storing init CLI args.
    """

    source_host: str
    dest_host: str
    indexes: Optional[List[str]]
    concurrent_tasks: int = DEFAULT_CONCURRENT_TASKS
    check_interval: int = DEFAULT_CHECK_INTERVAL



from typing import List

from elasticsearch import Elasticsearch, exceptions

from elasticsearch_reindex.const import DEFAULT_ES_KWARGS
from elasticsearch_reindex.errors import (
    ES_NODE_NOT_FOUND_ERROR,
    ElasticSearchNodeNotFoundException,
)
from elasticsearch_reindex.schema import Index
from elasticsearch_reindex.utils import chunkify


class ElasticsearchClient:
    """
    Client for manipulation with Elasticsearch clients and indexes.
    """

    def __init__(self, source_es_host: str, dest_es_host: str) -> None:
        self.source_es_host = source_es_host
        self.dest_es_host = dest_es_host

    @property
    def source_client(self) -> Elasticsearch:
        """
        Return Elasticsearch client where data will be transferred from.
        """
        return self._get_es_client(es_host=self.source_es_host)

    @property
    def dest_client(self) -> Elasticsearch:
        """
        Return Elasticsearch client where data will be transferred.
        """
        return self._get_es_client(es_host=self.dest_es_host)

    def get_source_indexes(self) -> List[Index]:
        """
        Return all indexes in Elasticsearch and amount of documents.
        """
        indexes = self.source_client.cat.indices(h="index,docs.count", s="index")
        return self._get_all_indexes(indexes=indexes.split())

    def get_dest_indexes(self) -> List[Index]:
        """
        Return all indexes in Elasticsearch and amount of documents.
        """
        indexes = self.dest_client.cat.indices(h="index,docs.count", s="index")
        return self._get_all_indexes(indexes=indexes.split())

    @staticmethod
    def get_user_indexes(
        source_indexes: List[Index], user_indexes: List[str]
    ) -> List[Index]:
        """
        Compare indexes provided by user.
        Return indexes for migration.

        :param source_indexes: List of source indexes.
        :param user_indexes: List of user indexes.
        :return: List of `Index` objects.
        """
        return [index for index in source_indexes if index.name in set(user_indexes)]

    @staticmethod
    def _get_es_client(es_host: str) -> Elasticsearch:
        """
        Ping ElasticSearch server and return initialized client object.
        """
        client = Elasticsearch(hosts=es_host, **DEFAULT_ES_KWARGS)  # type: ignore
        try:
            client.info()
        except exceptions.ConnectionError:
            raise ElasticSearchNodeNotFoundException(
                message=ES_NODE_NOT_FOUND_ERROR.format(host=es_host)
            )
        return client

    @staticmethod
    def _get_all_indexes(indexes: List[str]) -> List[Index]:
        """
        Return all indexes in Elasticsearch and amount of documents.
        """
        return [
            Index(name=name, docs_count=int(count))
            for name, count in chunkify(lst=indexes, n=2)
            if not name.startswith(".")
        ]



from datetime import datetime
from time import sleep
from typing import Generator, List

import pytest

from elasticsearch_reindex.client import ElasticsearchClient
from elasticsearch_reindex.manager import Manager
from elasticsearch_reindex.reindex import ReindexService
from elasticsearch_reindex.utils import bulk_insert
from tests.config import ES_TEST_INDEX, LOCAL_DEST_HOST, LOCAL_SOURCE_HOST, TEST_ENV

TEST_INDEXES_COUNT = 3
DOCS_COUNT = 10000

DEFAULT_INPUT_PARAMS = {
    "source_host": LOCAL_SOURCE_HOST,
    "dest_host": LOCAL_DEST_HOST,
    "check_interval": 10,
    "concurrent_tasks": 2,
}


@pytest.fixture(autouse=True)
def check_testing_mode():
    """
    Check that testing env was set before each test.
    """
    assert TEST_ENV is True
    yield


@pytest.fixture()
def test_messages() -> List[dict]:
    """
    Return messages with different id.
    """
    return [
        {
            "date": datetime.strptime("August 3, 2021", "%B %d, %Y").isoformat(),
            "id": message_id ** 2,
            "text": "Some review",
            "source": "amazon",
        }
        for message_id in range(DOCS_COUNT)
    ]


@pytest.fixture()
def reindex_service() -> ReindexService:
    """
    Return initialized reindex service.
    """
    reindex_service = ReindexService(
        source_es_host=LOCAL_SOURCE_HOST, dest_es_host=LOCAL_DEST_HOST
    )
    return reindex_service


@pytest.fixture()
def elastic_client() -> ElasticsearchClient:
    """
    Return initialized reindex service.
    """
    es_client = ElasticsearchClient(
        source_es_host=LOCAL_SOURCE_HOST, dest_es_host=LOCAL_DEST_HOST
    )
    return es_client


@pytest.fixture()
def insert_indexes(
    test_messages: List[dict], elastic_client: ElasticsearchClient
) -> Generator[ReindexService, None, None]:
    """
    Return reindex service.
    """
    for index_number in range(TEST_INDEXES_COUNT):
        es_index = ES_TEST_INDEX.format(number=index_number)
        bulk_insert(
            docs=test_messages,
            es_client=elastic_client.source_client,
            es_index=es_index,
        )
    # Wait some time when data will reach Elasticsearch.
    sleep(2)
    yield elastic_client


@pytest.fixture()
def manager(elastic_client: ElasticsearchClient) -> Generator[Manager, None, None]:
    """
    Return initialized manager class for start reindexing.
    """
    manager_class = Manager.from_dict(data=DEFAULT_INPUT_PARAMS)
    yield manager_class

    # Clear all of test indexes from source ans destination Elastics.
    for index_number in range(TEST_INDEXES_COUNT):
        index = ES_TEST_INDEX.format(number=index_number)
        elastic_client.source_client.indices.delete(index=index)
        elastic_client.dest_client.indices.delete(index=index)



"""
Provide errors for project.
"""
ES_NODE_NOT_FOUND_ERROR = "Can not connect to ElasticSearch server: {host}"
ES_TASK_ID_ERROR = (
    "Can not retrieve task status "
    "from ElasticSearch server: {host} and task id: {task_id}"
)


class BaseCustomException(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class ElasticSearchNodeNotFoundException(BaseCustomException):
    """
    Exception raised when Elasticsearch node not found or can not get access.
    """


class ElasticSearchInvalidTaskIDException(BaseCustomException):
    """
    Exception raised when got ElasticSearch invalid task ID.
    """



"""
Variables for testing.
"""
import os

TEST_ENV = os.getenv("ENV") == "test"

ES_SOURCE_HOST = os.getenv("LOCAL_IP")
ES_SOURCE_PORT = os.getenv("ES_SOURCE_PORT")
ES_DEST_PORT = os.getenv("ES_DEST_PORT")

LOCAL_SOURCE_HOST = f"http://{ES_SOURCE_HOST}:{ES_SOURCE_PORT}"
LOCAL_DEST_HOST = f"http://{ES_SOURCE_HOST}:{ES_DEST_PORT}"

# Elasticsearch test index.
ES_TEST_INDEX = "test_index_{number}"



import logging

from elasticsearch_reindex.const import LOG_FORMAT


def create_logger(logger_name: str) -> logging.Logger:
    """
    Initialize logger for project.
    """
    logger = logging.getLogger(logger_name)

    logging.basicConfig(format=LOG_FORMAT)
    logger.setLevel(level=logging.INFO)

    logging.getLogger("urllib3").setLevel(logging.ERROR)
    logging.getLogger("elasticsearch").setLevel(logging.ERROR)

    return logger



from typing import List

import click
from click_default_group import DefaultGroup

from elasticsearch_reindex.manager import Manager


@click.group(cls=DefaultGroup, default="reindex", default_if_no_args=True)
def cli() -> None:
    """
    Main click group.
    """


@cli.command()
@click.option(
    "--source_host",
    required=True,
    type=str,
    help="Source server: Elasticsearch host where data will be transferred from",
)
@click.option(
    "--dest_host",
    required=True,
    type=str,
    help="Destination server: Elasticsearch host where data will be transferred",
)
@click.option(
    "--check_interval",
    required=False,
    type=int,
    help="Interval for check Elasticsearch reindex task (in seconds)",
)
@click.option(
    "--concurrent_tasks",
    required=False,
    type=int,
    help="Number of max concurrent reindex tasks",
)
@click.option(
    "--indexes",
    "-i",
    required=True,
    multiple=True,
    help="List of specific Elasticsearch indexes to migrate",
)
def reindex(
    source_host: str,
    dest_host: str,
    check_interval: int,
    concurrent_tasks: int,
    indexes: List[str],
) -> None:
    """
    Upload JSON data files from Github to MongoDB.
    """
    config = {
        "source_host": source_host,
        "dest_host": dest_host,
        "check_interval": check_interval,
        "concurrent_tasks": concurrent_tasks,
        "indexes": list(indexes),
    }
    reindex_manager = Manager.from_dict(data=config)
    reindex_manager.start_reindex()
    click.echo("Success")



import os
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from typing import Dict

from elasticsearch_reindex.client import ElasticsearchClient
from elasticsearch_reindex.const import DEFAULT_CHECK_INTERVAL, DEFAULT_CONCURRENT_TASKS
from elasticsearch_reindex.logs import create_logger
from elasticsearch_reindex.reindex import ReindexService
from elasticsearch_reindex.schema import Config
from elasticsearch_reindex.utils import check_migrated_indexes

logger = create_logger(__name__)


class Manager:
    """
    Logic for input args handling.
    """

    def __init__(self, config: Config) -> None:
        self.config = config
        self.es_client = ElasticsearchClient(
            source_es_host=config.source_host, dest_es_host=config.dest_host
        )
        self.reindex_service = ReindexService(
            source_es_host=config.source_host, dest_es_host=config.dest_host
        )

    @classmethod
    def from_dict(cls, data: dict) -> "Manager":
        """
        Initialize Manages class from dict settings.
        """
        config = Config(
            source_host=data["source_host"],
            dest_host=data["dest_host"],
            indexes=data.get("indexes", []),
            check_interval=data.get("check_interval", DEFAULT_CHECK_INTERVAL),
            concurrent_tasks=data.get("concurrent_tasks", DEFAULT_CONCURRENT_TASKS),
        )
        return cls(config=config)

    def start_reindex(self) -> None:
        """
        Branching logic by input action.
        """
        source_indexes = self.es_client.get_source_indexes()
        if self.config.indexes:
            source_indexes = self.es_client.get_user_indexes(
                user_indexes=self.config.indexes, source_indexes=source_indexes
            )
        logger.info(f"Source ES host has {len(source_indexes)} indexes")

        dest_indexes = self.es_client.get_dest_indexes()
        logger.info(f"Destination ES host has {len(dest_indexes)} indexes")

        not_migrated_indexes, partial_migrated_indexes = check_migrated_indexes(
            source_indexes=source_indexes, dest_indexes=dest_indexes
        )
        logger.info(
            f"Not migrated ES indexes: {len(not_migrated_indexes)}/{len(source_indexes)}"
        )
        logger.info(
            f"Partial migrated ES indexes: {len(partial_migrated_indexes)}/{len(source_indexes)}"
        )

        # Calculate max concurrent task depends on CPU count.
        max_workers = min(self.config.concurrent_tasks, (os.cpu_count() or 1) * 5)
        executor = ThreadPoolExecutor(max_workers=max_workers)

        futures = {}
        for es_index in not_migrated_indexes:
            kwargs = {
                "es_index": es_index,
                "source_es_host": self.config.source_host,
                "dest_es_host": self.config.dest_host,
                "check_interval": self.config.check_interval,
            }
            future = executor.submit(self.reindex_service.transfer_index, **kwargs)
            futures[future] = es_index

        self._process_result(futures=futures)

    @staticmethod
    def _process_result(futures: Dict[Future, str]) -> None:
        """
        Process ThreadPoolExecutor tasks result.
        """
        for future in as_completed(futures):
            index = futures[future]
            try:
                task_id = future.result()
                futures.pop(future, None)
            except Exception as exc:
                logger.error(f"Index: {index} generated an exception: {exc}")
            else:
                logger.info(f"Task id: {task_id}. Reindex completed: {index}.")
                logger.info(f"Tasks left: {len(futures)}")



# Default Headers for call ElasticSearch API.
HEADERS = {"Content-Type": "application/json"}

# Default arguments for ElasticSearch client init.
DEFAULT_ES_KWARGS = {"max_retries": 3, "timeout": 30, "retry_on_timeout": False}

# Endpoint for create internal ElasticSearch reindex task.
ES_CREATE_TASK_ENDPOINT = "{es_host}/_reindex?pretty&wait_for_completion=false"
ES_CHECK_TASK_ENDPOINT = "{es_host}/_tasks/{task_id}"

# Logging message format.
LOG_FORMAT = "[%(asctime)s] %(message)s"

DEFAULT_CHECK_INTERVAL = 10
DEFAULT_CONCURRENT_TASKS = 1



from typing import Iterable, List, Tuple

from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

from elasticsearch_reindex.logs import create_logger
from elasticsearch_reindex.schema import Index

logger = create_logger(__name__)


def chunkify(lst: List, n: int) -> Iterable:
    """
    Yield successive n-sized chunks from list.
    """
    # Slice needs for bypass black failing.
    yield from (lst[slice(i, i + n)] for i in range(0, len(lst), n))


def _get_insert_action(body: dict, es_index: str) -> dict:
    return {"_index": es_index, **body}


def bulk_insert(docs: List[dict], es_client: Elasticsearch, es_index: str) -> None:
    """
    Execute insert actions to ElasticSearch.

    :param docs: List with document to insert.
    :param es_client: ElasticSearch initialized client.
    :param es_index: ElasticSearch index where data will be inserted.
    """
    actions = [_get_insert_action(body=item, es_index=es_index) for item in docs]
    success, failed = bulk(client=es_client, actions=actions, stats_only=True)
    logger.info(f"Successfully actions: {success}. Failed actions: {failed}")


def get_flatten_dict(data: List[Index]) -> dict:
    """
    Convert list of dataclasses to dict for fast searching.
    """
    return {index.name: index.docs_count for index in data}


def check_migrated_indexes(
    source_indexes: List[Index], dest_indexes: List[Index]
) -> Tuple[list, list]:
    """
    Check if index from `source_indexes` exist in `dest_indexes`.
    If index already exist we should check if all documents was transferred.
    """
    source_indexes = get_flatten_dict(data=source_indexes)
    dest_indexes = get_flatten_dict(data=dest_indexes)

    not_migrated, partial_migrated = [], []

    for index, docs_count in source_indexes.items():
        if dest_indexes.get(index):
            if docs_count != dest_indexes[index]:
                partial_migrated.append(index)
        else:
            not_migrated.append(index)

    return not_migrated, partial_migrated



from typing import List

import pytest
from elasticsearch import Elasticsearch

from elasticsearch_reindex import Manager
from elasticsearch_reindex.client import ElasticsearchClient
from elasticsearch_reindex.errors import ElasticSearchNodeNotFoundException
from elasticsearch_reindex.schema import Index
from tests.conftest import TEST_INDEXES_COUNT

INVALID_SOURCE_HOST = "localhost:9900"
INVALID_DEST_HOST = "localhost:9901"


def test_invalid_es_connection() -> None:
    """
    Test valid Elasticsearch client initialization.
    """
    es_client = ElasticsearchClient(
        source_es_host=INVALID_SOURCE_HOST, dest_es_host=INVALID_DEST_HOST
    )
    with pytest.raises(ElasticSearchNodeNotFoundException):
        assert isinstance(es_client.source_client, Elasticsearch)

    with pytest.raises(ElasticSearchNodeNotFoundException):
        assert isinstance(es_client.dest_client, Elasticsearch)


def test_valid_es_connection(elastic_client: ElasticsearchClient) -> None:
    """
    Test valid Elasticsearch client initialization.
    """
    assert isinstance(elastic_client.source_client, Elasticsearch)
    assert isinstance(elastic_client.dest_client, Elasticsearch)


def test_valid_indexes_creation(
    insert_indexes, elastic_client: ElasticsearchClient
) -> None:
    """
    Test valid indexes creation.
    """
    data: List[Index] = elastic_client.get_source_indexes()
    assert len(data) == TEST_INDEXES_COUNT


def test_valid_data_migration(
    manager: Manager, elastic_client: ElasticsearchClient
) -> None:
    """
    Test successfully reindexing.
    """
    manager.start_reindex()
    indexes = elastic_client.get_dest_indexes()
    assert len(indexes) == TEST_INDEXES_COUNT



"""
Setup the package.
"""
from setuptools import find_packages, setup

project_name = "elasticsearch-reindex"
package_name = "elasticsearch_reindex"

description = (
    "Python package for simple migration elasticsearch indexes between servers."
)
version = "1.2.0"

with open("README.md") as read_me:
    long_description = read_me.read()

packages = [package for package in find_packages(where=".", exclude=("test*",))]

install_requires = [
    "click-default-group==1.2.2",
    "elasticsearch>=7.13.4,<8",
    "requests>=2.26.0",
]

setup(
    version=version,
    name=project_name,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/borys25ol/elasticsearch-reindex",
    license="MIT",
    entry_points="""
       [console_scripts]
       elasticsearch_reindex=elasticsearch_reindex.__main__:cli
    """,
    author="Borys Oliinyk",
    author_email="oleynik.boris@gmail.com",
    packages=packages,
    package_data={package_name: ["py.typed"]},
    include_package_data=True,
    install_requires=install_requires,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



from time import sleep
from typing import Dict, Tuple, Union

import requests

from elasticsearch_reindex.client import ElasticsearchClient
from elasticsearch_reindex.const import (
    ES_CHECK_TASK_ENDPOINT,
    ES_CREATE_TASK_ENDPOINT,
    HEADERS,
)
from elasticsearch_reindex.errors import (
    ES_TASK_ID_ERROR,
    ElasticSearchInvalidTaskIDException,
)
from elasticsearch_reindex.logs import create_logger

logger = create_logger(__name__)


class ReindexService:
    """
    This class provide simple interface to ElasticSearch reindex API.
    """

    def __init__(self, source_es_host: str, dest_es_host: str):
        self.source_es_host = source_es_host
        self.dest_es_host = dest_es_host
        self._es_service = ElasticsearchClient(
            source_es_host=source_es_host, dest_es_host=dest_es_host
        )

    def transfer_index(
        self,
        es_index: str,
        source_es_host: str,
        dest_es_host: str,
        check_interval: int = 10,
    ) -> str:
        """
        Create reindex task and waiting for finish this task.

        :param es_index: Elasticsearch index.
        :param source_es_host: Source Elasticsearch host.
        :param dest_es_host: Destination Elasticsearch host.
        :param check_interval: Period of request task status (in seconds).
        """
        task_id = self._create_reindex_task(
            es_index=es_index, source_es_host=source_es_host, dest_es_host=dest_es_host
        )
        logger.info(f"Got task for migrate data: {task_id}")

        # Wait for Elasticsearch manage input task.
        sleep(2)

        while True:
            completed, status = self._check_task_completed(
                dest_es_host=self.dest_es_host, task_id=task_id
            )
            logger.info(
                f"Task id: {task_id}. "
                f"Migrated documents: {status['created']}/{status['total']}"
            )
            if not completed:
                sleep(check_interval)
                continue

            logger.info(f"Task finished: {task_id}")
            break

        return task_id

    def _create_reindex_task(
        self, es_index: str, source_es_host: str, dest_es_host: str
    ) -> str:
        """
        Create reindex task via Elasticsearch API.
        """
        endpoint = ES_CREATE_TASK_ENDPOINT.format(es_host=dest_es_host)
        reindex_payload = self._get_reindex_body(
            es_index=es_index, source_es_host=source_es_host
        )
        response = requests.post(url=endpoint, json=reindex_payload, headers=HEADERS)
        task_id = response.json()["task"]
        return task_id

    @staticmethod
    def _check_task_completed(
        dest_es_host: str, task_id: str
    ) -> Tuple[bool, Dict[str, int]]:
        """
        Make request to Elasticsearch Tasks API and check task status.
        """
        endpoint = ES_CHECK_TASK_ENDPOINT.format(es_host=dest_es_host, task_id=task_id)
        response = requests.get(url=endpoint).json()

        if (
            "error" in response
            and response["error"]["type"] == "illegal_argument_exception"
        ):
            raise ElasticSearchInvalidTaskIDException(
                message=ES_TASK_ID_ERROR.format(host=dest_es_host, task_id=task_id)
            )

        response_status = response["task"]["status"]

        data = {
            "total": response_status["total"],
            "created": response_status["created"],
        }
        return response["completed"], data

    @staticmethod
    def _get_reindex_body(
        es_index: str, source_es_host: str
    ) -> Dict[str, Union[str, dict]]:
        """
        Return ElasticSearch reindex body for API request.
        """
        return {
            "source": {"remote": {"host": source_es_host}, "index": es_index},
            "conflicts": "proceed",
            "dest": {"index": es_index},
        }
